import anndata as ad
import numpy as np


def guess_is_log(adata: ad.AnnData, num_cells: int | float = 5e2) -> bool:
    """Make an *educated* guess whether the provided anndata is log-transformed.

    Selects a random subset of cells and sums their counts.
    Returns false if all decimal components are zero (unlikely for log transformed data)
    """
    # Select either the provided `num_cells` or the maximum number of cells in the `adata`
    num_cells = int(min(num_cells, adata.shape[0]))

    # Draw a random mask of cells
    mask = np.random.choice(adata.shape[0], size=num_cells, replace=False)

    # Sum the matrix across the selected cell subset
    sums = adata[mask].X.sum(axis=1)

    # Extract the fractional components of the array
    decimals, _ = np.modf(sums)

    return np.any(decimals != 0.0)



import anndata as ad
import numpy as np

from pdex._utils import guess_is_log

N_CELLS = 1000
N_GENES = 10
MAX_COUNT = 1e6


def build_anndata(log=False) -> ad.AnnData:
    dim = (N_CELLS, N_GENES)
    return ad.AnnData(
        X=np.random.random(size=dim)
        if log
        else np.random.randint(0, MAX_COUNT, size=dim)
    )


def test_log_guess():
    log_anndata = build_anndata(log=True)
    assert guess_is_log(log_anndata)

    count_anndata = build_anndata(log=False)
    assert not guess_is_log(count_anndata)



import anndata as ad
import pandas as pd
import polars as pl
from adpbulk import ADPBulk
from pydeseq2.dds import DeseqDataSet
from pydeseq2.default_inference import DefaultInference
from pydeseq2.ds import DeseqStats
from tqdm import tqdm

KNOWN_METHODS = ["sum", "mean", "median"]


def pseudobulk_dex(
    adata: ad.AnnData,
    groupby: list[str],
    groups: list[str] | None = None,
    test_col: str = "target_gene",
    reference: str = "non-targeting",
    design: str | None = None,
    num_workers: int = 1,
    aggr_method: str = "sum",
    as_polars: bool = False,
) -> pd.DataFrame | pl.DataFrame:
    """Calculate differential expression between groups of cells after performing a pseudobulk.


    Parameters
    ----------
    adata: ad.AnnData
        Annotated data matrix containing raw gene expression data (count data)
    groupby: list[str]
        Which columns to pseudobulk over - must include `test_col` (perturbation column)
    groups: list[str], optional
        List of targets to compare, defaults to None which compares all groups (all perturbations)
    reference: str, optional
        Reference group to compare against, defaults to "non-targeting"
    test_col: str, optional
        Column in obs which designates the perturbation identity (or what you want to measure differential expression)
    design: str, optional
        Design formula to use in DESeq2 - defaults to "~ test_col + (remaining groupby keys)"
    num_workers: int
        Number of workers to use for parallel processing, defaults to 1
    aggr_method: str
        Aggregation method to use in pseudobulk
    as_polars: bool
        Return the dataframe as a polars dataframe

    Returns
    -------
    pd.DataFrame containing differential expression results for each group and feature
    """

    # Check for known aggregation method
    if aggr_method not in KNOWN_METHODS:
        raise ValueError(
            f"Unknown aggr method: {aggr_method} :: Expecting: {KNOWN_METHODS} "
        )

    # Validate the test column is in groupby
    if test_col not in groupby:
        raise ValueError(
            f"The test column: {test_col} must be in the groupby set: {groupby}"
        )

    if len(groupby) < 2:
        raise ValueError(
            "DESeq2 requires replicates to properly estimate variance - you will need to provide an additional groupby "
            "column alongside the test column to pseudobulk across."
        )

    if design is None:
        design = "~" + "+".join(groupby)

    print(f"DESIGN: {design}")

    if groups is None:
        groups = [x for x in adata.obs[test_col].unique() if x != reference]

    bulk = ADPBulk(adata, groupby=groupby, method=aggr_method)
    matrix = bulk.fit_transform()
    meta = bulk.get_meta()

    inference = DefaultInference(n_cpus=num_workers)
    dds = DeseqDataSet(
        counts=matrix,
        metadata=meta.set_index("SampleName"),
        design=design,
        refit_cooks=True,
        inference=inference,
        quiet=True,
    )
    dds.deseq2()

    all_results = []
    for target in tqdm(groups, desc="Evaluating contrasts..."):
        ds = DeseqStats(
            dds,
            contrast=[test_col, target, reference],
            inference=inference,
            quiet=True,
        )
        ds.summary()
        results = ds.results_df.reset_index(names="gene")
        results["target"] = target
        results["reference"] = reference
        all_results.append(results)

    all_results = pd.concat(all_results)

    if as_polars:
        return pl.DataFrame(all_results)

    return all_results



import anndata as ad
import numpy as np
import pandas as pd
import polars as pl
from scipy.sparse import csr_matrix

from pdex import parallel_differential_expression

PERT_COL = "perturbation"
CONTROL_VAR = "control"

N_CELLS = 1000
N_GENES = 100
N_PERTS = 10
MAX_UMI = 1e6

RANDOM_SEED = 42


def build_random_anndata(
    n_cells: int = N_CELLS,
    n_genes: int = N_GENES,
    n_perts: int = N_PERTS,
    pert_col: str = PERT_COL,
    control_var: str = CONTROL_VAR,
    random_state: int = RANDOM_SEED,
) -> ad.AnnData:
    """Sample a random AnnData object."""
    if random_state is not None:
        np.random.seed(random_state)
    return ad.AnnData(
        X=np.random.randint(0, MAX_UMI, size=(n_cells, n_genes)),
        obs=pd.DataFrame(
            {
                pert_col: np.random.choice(
                    [f"pert_{i}" for i in range(n_perts)] + [control_var],
                    size=n_cells,
                    replace=True,
                ),
            },
            index=np.arange(n_cells).astype(str),
        ),
    )


def test_dex_dense_array():
    adata = build_random_anndata()
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_dense_array_log():
    adata = build_random_anndata()
    adata.X = np.log1p(adata.X)
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_dense_matrix():
    adata = build_random_anndata()
    adata.X = np.matrix(adata.X)
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_sparse_matrix():
    adata = build_random_anndata()
    adata.X = csr_matrix(adata.X)
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_wilcoxon():
    adata = build_random_anndata()
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
        metric="wilcoxon",
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_wilcoxon_no_tie_correct():
    adata = build_random_anndata()
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
        metric="wilcoxon",
        tie_correct=False,
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_anderson():
    adata = build_random_anndata()
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
        metric="anderson",
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_ttest():
    adata = build_random_anndata()
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
        metric="t-test",
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_dex_unknown_metric():
    adata = build_random_anndata()
    try:
        parallel_differential_expression(
            adata,
            reference=CONTROL_VAR,
            groupby_key=PERT_COL,
            metric="unknown",
        )
        assert False
    except ValueError:
        "Caught error"


def test_dex_single_observed_value_anderson():
    adata = build_random_anndata()
    adata.X = np.zeros_like(adata.X)
    parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
        metric="anderson",
    )


def test_dex_polars_output():
    adata = build_random_anndata()
    results = parallel_differential_expression(
        adata,
        reference=CONTROL_VAR,
        groupby_key=PERT_COL,
        as_polars=True,
    )
    assert results.shape[0] == N_GENES * N_PERTS
    assert isinstance(results, pl.DataFrame)



import logging
import multiprocessing as mp
from collections.abc import Iterator
from functools import partial
from multiprocessing.shared_memory import SharedMemory

import anndata as ad
import numpy as np
import pandas as pd
import polars as pl
from adjustpy import adjust
from scipy.sparse import csr_matrix
from scipy.stats import anderson_ksamp, mannwhitneyu, ttest_ind
from tqdm import tqdm

from ._utils import guess_is_log

# Configure logger
tools_logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

KNOWN_METRICS = ["wilcoxon", "anderson", "t-test"]


def _build_shared_matrix(
    data: np.ndarray | np.matrix | csr_matrix,
) -> tuple[str, tuple[int, int], np.dtype]:
    """Create a shared memory matrix from a numpy array."""
    if isinstance(data, np.matrix):
        data = np.asarray(data)
    elif isinstance(data, csr_matrix):
        data = data.toarray()
    shared_matrix = SharedMemory(create=True, size=data.nbytes)
    matrix = np.ndarray(data.shape, dtype=data.dtype, buffer=shared_matrix.buf)
    matrix[:] = data
    return shared_matrix.name, data.shape, data.dtype


def _conclude_shared_memory(name: str):
    """Close and unlink a shared memory."""
    shm = SharedMemory(name=name)
    shm.close()
    shm.unlink()


def _combinations_generator(
    target_masks: dict[str, np.ndarray],
    var_indices: dict[str, int],
    reference: str,
    target_list: list[str],
    feature_list: list[str],
) -> Iterator[tuple]:
    """Generate all combinations of target genes and features."""
    for target in target_list:
        for feature in feature_list:
            yield (
                target_masks[target],
                target_masks[reference],
                var_indices[feature],
                target,
                reference,
                feature,
            )


def _batch_generator(
    combinations: Iterator[tuple],
    batch_size: int,
    num_combinations: int,
) -> Iterator[list[tuple]]:
    """Generate batches of combinations."""
    for _i in range(0, num_combinations, batch_size):
        subset = []
        for _ in range(batch_size):
            try:
                subset.append(next(combinations))
            except StopIteration:
                break
        yield subset


def _process_target_batch_shm(
    batch_tasks: list[tuple],
    shm_name: str,
    shape: tuple[int, int],
    dtype: np.dtype,
    metric: str,
    tie_correct: bool = False,
    is_log1p: bool = False,
    **kwargs,
) -> list[dict[str, float]]:
    """Process a batch of target gene and feature combinations.

    This is the function that is parallelized across multiple workers.
    """
    # Open shared memory once for the batch
    existing_shm = SharedMemory(name=shm_name)
    matrix = np.ndarray(shape=shape, dtype=dtype, buffer=existing_shm.buf)

    results = []
    for (
        target_mask,
        reference_mask,
        var_index,
        target_name,
        reference_name,
        var_name,
    ) in batch_tasks:
        if target_name == reference_name:
            continue

        x_tgt = matrix[target_mask, var_index]
        x_ref = matrix[reference_mask, var_index]

        μ_tgt = _sample_mean(x_tgt, is_log1p=is_log1p)
        μ_ref = _sample_mean(x_ref, is_log1p=is_log1p)

        fc = _fold_change(μ_tgt, μ_ref)
        pcc = _percent_change(μ_tgt, μ_ref)

        (pval, stat) = (1.0, np.nan)  # default output in case of failure
        try:
            match metric:
                case "wilcoxon":
                    if tie_correct:
                        # default mannwhitneyu behavior
                        de_result = mannwhitneyu(
                            x_tgt, x_ref, use_continuity=True, **kwargs
                        )
                    else:
                        # equivalent to `ranksums` behavior when `use_continuity=False` but statistic changes
                        de_result = mannwhitneyu(
                            x_tgt, x_ref, use_continuity=False, **kwargs
                        )
                    pval, stat = (de_result.pvalue, de_result.statistic)
                case "anderson":
                    de_result = anderson_ksamp([x_tgt, x_ref], **kwargs)
                    pval, stat = (de_result.pvalue, de_result.statistic)
                case "t-test":
                    de_result = ttest_ind(x_tgt, x_ref, **kwargs)
                    pval, stat = (de_result.pvalue, de_result.statistic)
                case _:
                    raise KeyError(f"Unknown Metric: {metric}")
        except ValueError:
            """Don't bail on runtime value errors - just use default values"""

        results.append(
            {
                "target": target_name,
                "reference": reference_name,
                "feature": var_name,
                "target_mean": μ_tgt,
                "reference_mean": μ_ref,
                "percent_change": pcc,
                "fold_change": fc,
                "p_value": pval,
                "statistic": stat,
            }
        )

    existing_shm.close()
    return results


def _get_obs_mask(
    adata: ad.AnnData,
    target_name: str,
    variable_name: str = "target_gene",
) -> np.ndarray:
    """Return a boolean mask for a specific target name in the obs variable."""
    return adata.obs[variable_name] == target_name


def _get_var_index(
    adata: ad.AnnData,
    target_gene: str,
) -> int:
    """Return the index of a specific gene in the var variable.

    Raises
    ------
    ValueError
        If the gene is not found in the dataset.
    """
    var_index = np.flatnonzero(adata.var.index == target_gene)
    if len(var_index) == 0:
        raise ValueError(f"Target gene {target_gene} not found in dataset")
    return var_index[0]


def _sample_mean(
    x: np.ndarray,
    is_log1p: bool,
) -> float:
    """Determine the sample mean of a 1D array.

    Exponenentiates and subtracts one if `is_log1p == True`
    """
    if is_log1p:
        return np.expm1(x).mean()
    else:
        return x.mean()


def _fold_change(
    μ_tgt: float,
    μ_ref: float,
) -> float:
    """Calculate the fold change between two means."""
    if μ_ref == 0:
        return np.nan
    return μ_tgt / μ_ref


def _percent_change(
    μ_tgt: float,
    μ_ref: float,
) -> float:
    """Calculate the percent change between two means."""
    if μ_ref == 0:
        return np.nan
    return (μ_tgt - μ_ref) / μ_ref


def parallel_differential_expression(
    adata: ad.AnnData,
    groups: list[str] | None = None,
    reference: str = "non-targeting",
    groupby_key: str = "target_gene",
    num_workers: int = 1,
    batch_size: int = 100,
    metric: str = "wilcoxon",
    tie_correct: bool = True,
    is_log1p: bool | None = None,
    as_polars: bool = False,
    **kwargs,
) -> pd.DataFrame | pl.DataFrame:
    """Calculate differential expression between groups of cells.

    Parameters
    ----------
    adata: ad.AnnData
        Annotated data matrix containing gene expression data
    groups: list[str], optional
        List of groups to compare, defaults to None which compares all groups
    reference: str, optional
        Reference group to compare against, defaults to "non-targeting"
    groupby_key: str, optional
        Key in `adata.obs` to group by, defaults to "target_gene"
    num_workers: int
        Number of workers to use for parallel processing, defaults to 1
    batch_size: int
        Number of combinations to process in each batch, defaults to 100
    metric: str
        The differential expression metric to use [wilcoxon, anderson, t-test]
    tie_correct: bool
        Whether to perform continuity (tie) correction for wilcoxon ranksum test
    is_log1p: bool, optional
        Specify exactly whether the data is log1p transformed - will use heuristic to check if not provided
        (see `pdex._utils.guess_is_log`).
    as_polars: bool
        return the output dataframe as a polars dataframe
    **kwargs:
        keyword arguments to pass to metric

    Returns
    -------
    pd.DataFrame containing differential expression results for each group and feature
    """
    if metric not in KNOWN_METRICS:
        raise ValueError(f"Unknown metric: {metric} :: Expecting: {KNOWN_METRICS}")

    unique_targets = adata.obs[groupby_key].unique()
    if groups is not None:
        unique_targets = [
            target
            for target in unique_targets
            if target in groups or target == reference
        ]
    unique_features = adata.var.index

    if not is_log1p:
        is_log1p = guess_is_log(adata)

    # Precompute the number of combinations and batches
    n_combinations = len(unique_targets) * len(unique_features)
    n_batches = n_combinations // batch_size + 1

    # Precompute masks for each target gene
    logger.info("Precomputing masks for each target gene")
    target_masks = {
        target: _get_obs_mask(
            adata=adata, target_name=target, variable_name=groupby_key
        )
        for target in tqdm(unique_targets, desc="Identifying target masks")
    }

    # Precompute variable index for each feature
    logger.info("Precomputing variable indices for each feature")
    var_indices = {
        feature: idx
        for idx, feature in enumerate(
            tqdm(unique_features, desc="Identifying variable indices")
        )
    }

    # Isolate the data matrix from the AnnData object
    logger.info("Creating shared memory memory matrix for parallel computing")
    (shm_name, shape, dtype) = _build_shared_matrix(data=adata.X)

    logger.info(f"Creating generator of all combinations: N={n_combinations}")
    combinations = _combinations_generator(
        target_masks=target_masks,
        var_indices=var_indices,
        reference=reference,
        target_list=unique_targets,
        feature_list=unique_features,
    )
    logger.info(f"Creating generator of all batches: N={n_batches}")
    batches = _batch_generator(
        combinations=combinations,
        batch_size=batch_size,
        num_combinations=n_combinations,
    )

    # Partial function for parallel processing
    task_fn = partial(
        _process_target_batch_shm,
        shm_name=shm_name,
        shape=shape,
        dtype=dtype,
        metric=metric,
        tie_correct=tie_correct,
        is_log1p=is_log1p,
        **kwargs,
    )

    logger.info("Initializing parallel processing pool")
    with mp.Pool(num_workers) as pool:
        logger.info("Processing batches")
        batch_results = list(
            tqdm(
                pool.imap(task_fn, batches),
                total=n_batches,
                desc="Processing batches",
            )
        )

    # Flatten results
    logger.info("Flattening results")
    results = [result for batch in batch_results for result in batch]

    # Close shared memory
    logger.info("Closing shared memory pool")
    _conclude_shared_memory(shm_name)

    dataframe = pd.DataFrame(results)
    dataframe["fdr"] = adjust(dataframe["p_value"].values, method="bh")

    if as_polars:
        return pl.DataFrame(dataframe)

    return dataframe



import anndata as ad
import numpy as np
import pandas as pd
import polars as pl

from pdex import pseudobulk_dex

PERT_COL = "perturbation"
CONTROL_VAR = "control"

N_CELLS = 1000
N_GENES = 100
N_PERTS = 10
MAX_UMI = 1e6
N_CONDITIONS = 4

RANDOM_SEED = 42


def build_random_anndata(
    n_cells: int = N_CELLS,
    n_genes: int = N_GENES,
    n_perts: int = N_PERTS,
    pert_col: str = PERT_COL,
    control_var: str = CONTROL_VAR,
    n_covariates: int = 0,
    covariate_magnitudes: list[int] = [],
    random_state: int = RANDOM_SEED,
) -> ad.AnnData:
    """Sample a random AnnData object."""
    if random_state is not None:
        np.random.seed(random_state)
    obs = pd.DataFrame(
        {
            pert_col: np.random.choice(
                [f"pert_{i}" for i in range(n_perts)] + [control_var],
                size=n_cells,
                replace=True,
            ),
        },
        index=np.arange(n_cells).astype(str),
    )

    for idx in np.arange(n_covariates):
        n_conditions = (
            N_CONDITIONS
            if len(covariate_magnitudes) == 0
            else covariate_magnitudes[idx]
        )
        obs[f"covar.{idx}"] = [
            f"enum.{i}" for i in np.random.choice(n_conditions, size=n_cells)
        ]

    return ad.AnnData(
        X=np.random.randint(0, MAX_UMI, size=(n_cells, n_genes)),
        obs=obs,
        var=pd.DataFrame(index=[f"gene.{j}" for j in np.arange(N_GENES)]),
    )


def test_pbdex_no_covar():
    adata = build_random_anndata()
    try:
        _ = pseudobulk_dex(
            adata, groupby=[PERT_COL], reference=CONTROL_VAR, test_col=PERT_COL
        )
        assert False, "Should not arrive here"
    except ValueError:
        assert True


def test_pbdex_empty_groupby():
    adata = build_random_anndata()
    try:
        _ = pseudobulk_dex(adata, groupby=[], reference=CONTROL_VAR, test_col=PERT_COL)
        assert False, "Should not arrive here"
    except ValueError:
        assert True


def test_pbdex_one_covar():
    adata = build_random_anndata(n_covariates=1)
    results = pseudobulk_dex(
        adata, reference=CONTROL_VAR, test_col=PERT_COL, groupby=[PERT_COL, "covar.0"]
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_pbdex_two_covar():
    adata = build_random_anndata(n_covariates=2)
    results = pseudobulk_dex(
        adata,
        reference=CONTROL_VAR,
        test_col=PERT_COL,
        groupby=[PERT_COL, "covar.0", "covar.1"],
    )
    assert results.shape[0] == N_GENES * N_PERTS


def test_pbdex_polars_output():
    adata = build_random_anndata(n_covariates=1)
    results = pseudobulk_dex(
        adata,
        reference=CONTROL_VAR,
        test_col=PERT_COL,
        groupby=[PERT_COL, "covar.0"],
        as_polars=True,
    )
    assert results.shape[0] == N_GENES * N_PERTS
    assert isinstance(results, pl.DataFrame)
